from django.core.checks import register, Tags, Error
from django.db import connections


PREFERED_COLLATIONS = (
    "utf8mb4_unicode_520_ci",
    "utf8mb4_unicode_520_ci",
    "utf8mb4_general_ci",
    "utf8mb4_unicode_ci",
)

E001_DB_COLLATE_MESSAGE = "Database %s uses a bad collation(%s)."
E001_DB_COLLATE_HINT = "ALTER DATABASE %s COLLATE %s;"
W001_DB_COLLATE_MESSAGE = "Database %s uses a valid, but not-preferred collation(%s)."
W001_DB_COLLATE_HINT = "ALTER DATABASE %s COLLATE %s;"

E002_TABLE_COLLATE_MESSAGE = "Table %s.%s uses a bad collation(%s)."
E002_TABLE_COLLATE_HINT = "USE %s; ALTER TABLE %s COLLATE %s;"
W002_TABLE_COLLATE_MESSAGE = (
    "Table %s.%s uses a valid, but not-preferred collation(%s)."
)
W002_TABLE_COLLATE_HINT = "USE %s; ALTER TABLE %s COLLATE %s;"


def get_utf8mb_collations(connection, cursor=None):
    with connection.cursor() as cursor:
        c.execute("SHOW COLLATION WHERE Charset='utf8mb4';")
        return frozenset([row[0] for row in c.fetchall()])


def get_preferred_collations(connection):
    valid_collations = get_utf8mb_collations(connection)
    return [c for c in PREFERED_COLLATIONS if c in valid_collations]


@register(Tags.database)
def check_utf8mb4_database(app_configs, **kwargs):
    errors = []
    for connection in connections.all():
        if connection.vendor != "mysql":
            continue
        db_name = connection.get_connection_params()["db"]
        valid_collations = get_utf8mb_collations(connection)
        preferred_collations = [c for c in PREFERED_COLLATIONS if c in valid_collations]
        with connection.cursor() as cursor:
            cursor.execute("SELECT @@collation_database;")
            collation, = cursor.fetchone()
        if collation not in valid_collations:
            errors.append(
                Error(
                    E001_DB_COLLATE_MESSAGE % (connection.alias, collation),
                    hint=E001_DB_COLLATE_HINT % (db_name, preferred_collations[0]),
                    id="tiaki.base.E001",
                )
            )
        elif collation not in preferred_collations:
            errors.append(
                Warning(
                    W001_DB_COLLATE_MESSAGE % (connection.alias, collation),
                    hint=W001_DB_COLLATE_HINT % (db_name, preferred_collations[0]),
                    id="tiaki.base.W001",
                )
            )
    return errors


@register(Tags.models)
def check_utf8mb_collation(app_configs, **kwargs):
    errors = []
    for connection in connections.all():
        if connection.vendor != "mysql":
            continue
        tables = connection.introspection.django_table_names()
        db_name = connection.get_connection_params()["db"]
        valid_collations = get_utf8mb_collations(connection)
        preferred_collations = [c for c in PREFERED_COLLATIONS if c in valid_collations]
        with connection.cursor() as cursor:
            cursor.execute(
                "SELECT TABLE_NAME, TABLE_COLLATION FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA=%s;",
                (db_name,),
            )
            table_collations = [
                (table, collation) for table, collation in cursor if table in tables
            ]
        for table, collation in table_collations:
            if collation not in valid_collations:
                errors.append(
                    Error(
                        E002_TABLE_COLLATE_MESSAGE
                        % (connection.alias, table, collation),
                        hint=E002_TABLE_COLLATE_HINT
                        % (db_name, table, preferred_collations[0]),
                        id="tiaki.base.E002",
                    )
                )
            elif collation not in preferred_collations:
                errors.append(
                    Warning(
                        W002_TABLE_COLLATE_MESSAGE
                        % (connection.alias, table, collation),
                        hint=W002_TABLE_COLLATE_HINT
                        % (db_name, table, preferred_collations[0]),
                        id="tiaki.base.W002",
                    )
                )
    return errors



from django_mysql_utf8mb4.checks import get_preferred_collations
from django.db import connections


def db_colation_migration(connection, cursor=None):
    if connection.vendor != "mysql":
        return
    collation = get_preferred_collations(connection)[0]
    db_name = connection.get_connection_params()["db"]
    sql = "ALTER DATABASE %s CHARACTER SET utfmb4 COLLATE %s;" % (db_name, collation)
    if cursor:
        cursor.execute(sql)
    else:
        with connection.cursor() as cursor:
            cursor.execute(sql)


def table_colation_migration(table_name, connection, cursor=None):
    if connection.vendor != "mysql":
        return
    collation = get_preferred_collations(connection)[0]
    sql = "ALTER TABLE %s CONVERT TO CHARACTER SET utf8mb4 COLLATE %s" % (
        table_name,
        collation,
    )
    if cursor:
        cursor.execute(sql)
    else:
        with connection.cursor() as cursor:
            cursor.execute(sql)


def django_tables_colation_migration(connection, cursor=None):
    if connection.vendor != "mysql":
        return
    for name in connection.introspection.django_table_names():
        table_colation_migration(name, connection, cursor=cursor)



from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.utils import OperationalError


class Command(BaseCommand):
    help = "Updates the database & table collations to the 'best' avaliable utf8mb4 collation"
    requires_migrations_checks = True

    def handle(self, *args, **kwargs):
        User = get_user_model()
        User.objects.filter(username=TEST_USERNAME).delete()
        try:
            user = User.objects.create_user(
                username=TEST_USERNAME,
                first_name="😀😀😀",
                last_name="😄😄😄",
                is_active=False,
            )
            self.stdout.write(
                self.style.SUCCESS(
                    "User with extended unicode characters created successfully"
                )
            )
        except OperationalError as e:
            if e.args[0] == 1366:
                raise CommandError(
                    "Failed to create user using 4-byte UTF-8. Check the database colation."
                ) from e
            else:
                raise CommandError(
                    "Another error occurred whilst creating the user:\n  %s" % e
                ) from e
        finally:
            User.objects.filter(username=TEST_USERNAME).delete()



from setuptools import find_packages, setup


packages = find_packages(
    where="./", include=["django_mysql_utf8mb4", "django_mysql_utf8mb4.*"]
)
if not packages:
    raise ValueError("No packages detected.")

setup(
    name="django_mysql_utf8mb4",
    version="0.1.1",
    description="A migration, command and checks to ensure Django and MySQL are using the best-avaliable collation",
    url="https://github.com/software-opal/django_mysql_utf8mb4",
    author="Opal Symes",
    author_email="pypi@opal.codes",
    python_requires=">=3.5",
    license="",
    packages=packages,
    install_requires=["django>=2.0"],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Security",
    ],
)



from django.db import migrations
from django_mysql_utf8mb4.migrate import (
    db_colation_migration,
    django_tables_colation_migration,
)


def forwards_db_colation_migration(apps, schema_editor):
    if schema_editor.connection.vendor != "mysql":
        return
    # schema_editor acts as a cursor.
    db_colation_migration(schema_editor.connection, cursor=schema_editor)


def forwards_table_colation_migration(apps, schema_editor):
    if schema_editor.connection.vendor != "mysql":
        return
    # schema_editor acts as a cursor.
    django_tables_colation_migration(schema_editor.connection, cursor=schema_editor)


class Migration(migrations.Migration):

    operations = [
        migrations.RunPython(forwards_db_colation_migration, atomic=False),
        migrations.RunPython(forwards_table_colation_migration, atomic=False),
    ]
